/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/*
 * This file contains the constants required be the caller of
 * a request.
 * the two make up the 2 dim array returned in the upcall.
 * char result[MAXREQUEST][RESULTLENGTH];
 *
 * $Log: req.h,v $
 * Revision 1.3  1996/08/16 19:06:17  tamches
 * updated copyright for release 1.1
 *
 * Revision 1.2  1994/05/04 00:24:00  hollings
 * Fixed pand and zoom to work. The conversion of the data type to double
 * broke many other things since all floats can become doubles!
 *
 * Make it pass through gcc -Wall with fewer errors.
 *
 * Revision 1.1  1994/03/31  18:10:06  krisna
 * interface files for the Histogram widget
 *
 * Revision 2.7  1991/06/28  16:28:09  rbi
 * onthefly derived metrics.
 *
 * Revision 2.6  1990/09/19  16:04:08  hollings
 * Support for lots of metrics using both multi-column and scrollbars.
 *
 * Revision 2.5  90/08/24  13:02:22  hollings
 * Upgrade to use formatted table widget and support option lists.
 * 
 * Revision 2.4  90/05/12  20:27:04  hollings
 * Bug fixes in display of K longest paths.
 * 
 * Revision 2.3  90/02/06  18:17:52  hollings
 * New Copyright
 * 
 * Revision 2.2  90/02/06  18:05:14  hollings
 * New Copyright
 * 
 * Revision 2.1  89/11/06  16:55:06  hollings
 * New copyright message
 * 
 * Revision 2.0  89/10/18  17:48:07  hollings
 * Toolkit Release
 * 
 * Revision 1.4  89/07/26  13:30:22  hollings
 * Assorted bug fixes
 * 
 * Revision 1.3  89/07/14  16:32:51  hollings
 * Increased max number of items to 25
 * 
 *
 */
#define MAX_OPTIONS	10	/* max options in a list */
#define RESULTLENGTH 	150  	/* the length of the string returned. */

#define REQ_STRING	0
#define REQ_BOOLEAN	1
#define REQ_EXCLUSIVE	2
#define REQ_ACTION	3	/* one of n action buttons */
#define REQ_LABEL       4

typedef struct _PropItem {
    int		type;			/* type of request */
    int		numOptions;		/* number of buttons options */
    char	*label;			/* label to place above it */
    char	*initValue;		/* inital value */
    char	*options[MAX_OPTIONS];	/* list of possible values */
    Widget	buttonW[MAX_OPTIONS];	/* list of possible values */
    char	value[RESULTLENGTH];	/* returned value */
    int         HLeft, HRight;          /* cols to highlight */
} PropItem;

typedef struct _PropertySheet {
    void	(*callBack)();		/* end callback */
    char	*title;			/* Title the window */
    int		buttonCount;		/* # of buttons */
    char	*buttons[MAX_OPTIONS];	/* buttons at bottom */
    int		itemCount;		/* number of items */
    int		type;			/* type of Request */
    int		cdata;			/* data to pass to callback func */
    PropItem	*items;			/* actual items */
} PropertySheet;

PropertySheet *InitPropertySheet();

#define RequestStringItem(sheet, slot, string, data)		\
    sheet->items[slot].type = REQ_STRING;			\
    sheet->items[slot].label = string;				\
    sheet->items[slot].initValue = data;			\
    sheet->items[slot].HLeft = -1;                              \
    sheet->items[slot].HRight = -1;                              \
    if (data) {							\
	strcpy(sheet->items[slot].value, data);			\
    }								\


#define RequestFromList(sheet, slot, string, count, data, def)		\
    sheet->items[slot].type = REQ_ACTION;				\
    sheet->items[slot].label = string;					\
    sheet->items[slot].numOptions = count;				\
    sheet->items[slot].initValue = data[def];				\
    {									\
	int i;								\
									\
	for (i=0; i < count; i++) {					\
	    sheet->items[slot].options[i] = data[i];			\
	}								\
    }									\

#define RequestLabelItem(sheet,slot,string)                     \
    sheet->items[slot].type = REQ_LABEL;                        \
    sheet->items[slot].label = string;

#ifdef __cplusplus
extern "C" {
#endif

void InitReqWindow(Widget w,
		   int x,
		   int y, 
		   int maxItemsPerCol,
		   char *icon_bits,
		   int icon_width,
		   int icon_height);

#ifdef __cplusplus
}
#endif
