/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/* 
 * $Id: FormatBox.c,v 1.3 1999/08/09 05:38:18 csserra Exp $
 * FormatBox.c - Formated box widget
 */

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Misc.h>
#include <X11/Xaw/XawInit.h>

#include "FormatBoxP.h"

/* Resources */
static XtResource resources[] = {
    { XtNformat, XtCFormat, XtRFormatOption, sizeof(int),
	XtOffsetOf(FormatBoxRec, format.justify_mode),
	XtRImmediate, (XtPointer) XtFormatNone },
    { XtNspanning, XtCSpanning, XtRBoolean, sizeof(Boolean),
	XtOffsetOf(FormatBoxRec, format.maximize),
	XtRImmediate, (XtPointer) False },
};

static void ChangedManaged(), InitStuff(), Resize();

FormatBoxClassRec formatBoxClassRec = {
  {
/* core_class fields      */
    /* superclass         */    (WidgetClass) &boxClassRec,
    /* class_name         */    "FormatedBox",
    /* widget_size        */    sizeof(FormatBoxRec),
    /* class_initialize   */    InitStuff,
    /* class_part_init    */	NULL,
    /* class_inited       */	FALSE,
    /* initialize         */    NULL,
    /* initialize_hook    */	NULL,
    /* realize            */    XtInheritRealize,
    /* actions            */    NULL,
    /* num_actions	  */	0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber(resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	TRUE,
    /* compress_exposure  */	TRUE,
    /* compress_enterleave*/	TRUE,
    /* visible_interest   */    FALSE,
    /* destroy            */    NULL,
    /* resize             */    Resize,
    /* expose             */    NULL,
    /* set_values         */    NULL,
    /* set_values_hook    */	NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */	NULL,
    /* accept_focus       */    NULL,
    /* version            */	XtVersion,
    /* callback_private   */    NULL,
    /* tm_table           */    NULL,
    /* query_geometry     */	(XtGeometryHandler) XtInheritQueryGeometry,
    /* display_accelerator*/	XtInheritDisplayAccelerator,
    /* extension          */	NULL
  },{
/* composite_class fields */
    /* geometry_manager   */    (XtGeometryHandler) XtInheritGeometryManager,
    /* change_managed     */    ChangedManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension          */	NULL
  },{
/* Box class fields */
    /* empty		  */	0,
  },{
/* Format Box class fields */
    /* empty              */    0,
  }
};

WidgetClass formatBoxWidgetClass = (WidgetClass)&formatBoxClassRec;

/* resource converter for formatting */
void StringToOption(XrmValue *args, int a_count, 
		    XrmValue *from, XrmValue *to)
{
    static int mode;
    char string[80];

    to->size = sizeof(int);
    to->addr = (caddr_t) &mode;
    if (!strcmp(from->addr, XtNleft)) {
	mode = XtFormatNone;
    } else if (!strcmp(from->addr, XtNcenter)) {
	mode = XtFormatCenter;
    } else if (!strcmp(from->addr, XtNpropSpace)) {
	mode = XtFormatJustify;
    } else {
	sprintf(string, "Invalid formatting option %s", from->addr);
	XtWarning(string);
	mode = XtFormatNone;
    }
}

static void InitStuff()
{
    XtAddConverter(XtRString, XtRFormatOption, (XtConverter) StringToOption, 
	NULL, 0);
}

/*
 * Guts of the format box widget.
 *
 */
static void Resize(FormatBoxWidget w)
{
    int h_space;
    char str[80];
    int x, extra;
    CoreWidget child;
    int i, width, height;

    if (w->format.maximize) {
	if (w->box.orientation == XtorientVertical) {
	    /* make width of children our width less spacing around edges */
	    width = w->core.width - 2 * w->box.h_space;
	    for (i = 0; i < w->composite.num_children; i++) {
		child = w->composite.children[i];
		XtResizeWidget(child, width - 2 * child->core.border_width, 
		    child->core.height, child->core.border_width);
	    }
	} else {
	    /* make height of children our height less spacing around edges */
	    height = w->core.height - 2 * w->box.v_space;
	    for (i = 0; i < w->composite.num_children; i++) {
		child = w->composite.children[i];
		XtResizeWidget(child, child->core.width, height, 
		    child->core.border_width);
	    }
	}
    }

    if (w->format.justify_mode == XtFormatNone) {
	return;
    } else {
	/* should add support for vertical justification */
	if (w->box.orientation == XtorientVertical) {
	    return;
	}

	/* compute used width */
	h_space = w->box.h_space;
	/* double count first h space */
	width = -h_space;
	for (i = 0; i < w->composite.num_children; i++) {
	    child = w->composite.children[i];
	    width += child->core.width + child->core.border_width * 2 + h_space;
	}

	extra = w->core.width - width;
	if (extra < 0) {
	    sprintf(str, "%s: Formated Box widget too small (size %d)",
		w->core.name, extra);
	    XtWarning(str);
	    return;
	}

	if (w->format.justify_mode == XtFormatCenter) {
	    extra /= 2;
	    x = extra;
	} else {
	    extra /= w->composite.num_children + 1;
	    x = extra;
	    h_space += extra;
	}

	/* now move the children */
	for (i = 0; i < w->composite.num_children; i++) {
	    child = w->composite.children[i];
	    XtMoveWidget(child, x, child->core.y);
	    x += child->core.width + child->core.border_width * 2 + h_space;
	}
    }
} 

/*
 * Call our superclass ChangedManaged function, and then force a call to resize
 *   to get the layout done to format the box.
 */
static void ChangedManaged(FormatBoxWidget w)
{
   CompositeClassRec *super= (CompositeClassRec *) 
       formatBoxClassRec.core_class.superclass;

   super->composite_class.change_managed((Widget) w);
   Resize(w);
}
